<?php
namespace Deargonaut\PaypalCsvToMt940;

use Deargonaut\PaypalCsvToMt940\Models\Mutation;
use Deargonaut\PaypalCsvToMt940\Models\Purpose;

class PaypalCsvToMt940
{
    private $csv_file_location;
    private $csv_array = [];
    private $options = [
        'filename'  => 'export.sta',
        'overwrite' => true,
        'location'  => './',
        'language'  => 'nl',
        'currency'  => 'EUR',
        'dateformat' => 'd-m-Y H:i:s e',
        'onlycompleted' => false
    ];

    private $parser;

    private $mutations = [

    ];


    public function __construct($csv, $language = 'NL', $options = [])
    {
	$this->setOptions($options);
        $this->loadFile($csv);
    }


    public function setDialect($dialect, ...$args)
    {
        $this->parser = new $dialect(...$args);
    }


    public function loadFile($csv)
    {
        if(!file_exists($csv))
            throw new \Exception("File [{$csv}] doesn't exist.");

        if(!\is_readable($csv))
            throw new \Exception("Can't open file. Check permissions.");

        $this->csv_file_location = $csv;

        // Remove the UTF-8 BOM flag in order to avoid wrong processing of
        // data. It seems, that str_getcsv is not able to UTF-8 BOM.
        $bom = pack('H*','EFBBBF');
        $text = preg_replace("/^$bom/", '', file_get_contents($this->csv_file_location));
        $csv_array = array_map(
            'str_getcsv', 
            explode("\n", $text)
        );

        array_walk($csv_array, function(&$a) use ($csv_array) {
            if (count($a) > 0 and !is_null($a[0]))
                $a = array_combine($csv_array[0], $a);
        });
        array_shift($csv_array); # remove column header
        $this->csv_array = $csv_array;
        $this->parse();

    }


    private function parse()
    {
        $f = include __DIR__ . '/HeaderTranslations.php';
        if(!isset($f[$this->options['language']]))
            $f = $f['nl'];
        else
            $f = $f[$this->options['language']];

        foreach($this->csv_array as $m)
        {
            // Skip empty lines.
            if (!$m) continue;
            
            if(!isset($m[$f['currency']]) || $m[$f['currency']] != $this->options['currency']) continue;
            // Skip pending entries, if requested.
            if($this->options['onlycompleted'] && $m[$f['status']] != $f['completed']) continue;
            // Skip entries, which are just a "memo", but not relevant.
            if($m[$f['effect_on_balance']] == $f['status_memo']) continue;

            $mutation = new Mutation();
            $mutation->setAmount($m[$f['gross']]);
            $mutation->setCurrency($m[$f['currency']]);
            $mutation->date = (\DateTime::createFromFormat($this->options['dateformat'], $m[$f['date']] . ' ' . 
                $m[$f['time']] . ' ' . $m[$f['timezone']]))->format('Y-m-d H:i:s e');
            $mutation->description = $m[$f['item_title']] . ' ' . $m[$f['type']] .  ' ' . 
                $m[$f['transaction_reference']] . ' ' . $m[$f['object_reference']] . ' ' . 
                $m[$f['invoice_number']] . ' ' . $m[$f['subject']];
            $mutation->setBalance($m[$f['balance']]);
            $mutation->setTransaction($m[$f['transaction_reference']]);
            $recipients = [];
            $recipientsMails = [];
            // credit = from
            // debit = to
            if (strlen($m[$f['name']]) > 0)
            {
                $recipients[] = $m[$f['name']];
            }
            if (strlen($m[$f['to_mail']]) > 0)
            {
                $recipientsMails[] = $m[$f['to_mail']];
            }
            if (strlen($m[$f['from_mail']]) > 0)
            {
                $recipientsMails[] = $m[$f['from_mail']];
            }
            if ($m[$f['effect_on_balance']] == $f['status_credit'])
            {
                $ppValue = 166;
                $recipientsMails = array_reverse($recipientsMails);
            }
            elseif ($m[$f['effect_on_balance']] == $f['status_debit'])
            {
                $ppValue = 116;
            }
            else
            {
                $ppValue = 835;
            }
            $mutation->addPurpose(new Purpose(Purpose::FLAG_CODE, $ppValue));
            if ($m[$f['status']] == $f['completed'])
            {
                $mutation->addPurpose(new Purpose(Purpose::FLAG_PRIMA, 'Completed'));
            }
            else
            {
                $mutation->addPurpose(new Purpose(Purpose::FLAG_PRIMA, 'Pending'));
            }
            $mutation->addPurpose(new Purpose(Purpose::FLAG_BOOK_TXT, 'Payment'));
            $recipients = array_merge($recipients, $recipientsMails);
            $mutation->addPurpose(Purpose::generateGeneric([Purpose::FLAG_NAME_1, Purpose::FLAG_NAME_2], $recipients));
            $mutation->addPurpose(Purpose::generatePurposeText($mutation->description));
            $this->mutations[] = $mutation;

            // In case a fee is indicated, we need to add additional mutation.
            if(Mutation::parseToFloat($m[$f['fee']]) > 0)
            {
                $mutation = new Mutation();
                $mutation->setAmount($m[$f['fee']]);
                $mutation->setCurrency($m[$f['currency']]);
                $mutation->date = (\DateTime::createFromFormat($this->options['dateformat'], $m[$f['date']] . ' ' . 
                    $m[$f['time']] . ' ' . $m[$f['timezone']]))->format('Y-m-d H:i:s e');
                $mutation->description = $f['fee'] . ' ' . $m[$f['item_title']] . ' ' . 
                    $m[$f['type']] .  ' ' . $m[$f['transaction_reference']] . ' ' . $m[$f['object_reference']] . ' ' . 
                    $m[$f['invoice_number']] . ' ' . $m[$f['subject']];
                $mutation->setBalance($m[$f['balance']]);
                $mutation->addPurpose(new Purpose(Purpose::FLAG_CODE, 808));
                $mutation->addPurpose(new Purpose(Purpose::FLAG_PRIMA, 'Completed'));
                $mutation->addPurpose(new Purpose(Purpose::FLAG_BOOK_TXT, $f['fee']));
                $mutation->addPurpose(new Purpose(Purpose::FLAG_NAME_1, 'Paypal'));
                $mutation->addPurpose(Purpose::generatePurposeText($mutation->description));
                $this->mutations[] = $mutation;
            }
        }
    }

    public function setOptions($options = [])
    {
        if(isset($options['filename']))
        {
            if(!is_string($options['filename']))
                throw new \Exception("Filename needs to be a string.");
            
            $fn = explode(DIRECTORY_SEPERATOR, $options['filename']);
            $this->options['filename'] = array_pop($fn);
        }

        if(isset($options['overwrite']))
        {
            if(!is_bool($options['overwrite']))
                throw new \Exception("Overwrite needs to be boolean");
            
            $this->options['overwrite'] = $options['overwrite'];
        }

        if(isset($options['location']))
        {
            if(!is_dir($options['location']))
                throw new \Exception("Location [{$options['location']}] is not a directory.");
            if(!\is_writeable($options['location']))
                throw new \Exception("Location is not writeable.");

            $this->options['location'] = $options['location'];
        }

        if(isset($options['language']))
        {
            $this->options['language'] = substr(strtolower($options['language']), 0, 2);
        }

        if(isset($options['currency']))
        {
            if(!is_string($options['currency']) || strlen($options['currency']) > 3)
                throw new \Exception("Invalid currency given as option.");
            $this->options['currency'] = $options['currency'];
        }

        if(isset($options['dateformat']))
        {
            if(!is_string($options['dateformat']))
                throw new \Exception("Invalid dateformat given.");
            $this->options['dateformat'] = $options['dateformat'];
        }

        if(isset($options['onlycompleted']))
        {
            if(!is_bool($options['onlycompleted']))
                throw new \Exception("Option \"onlycompleted\" must be boolean true/false!");
            $this->options['onlycompleted'] = $options['onlycompleted'];
        }
    }

    public function save($outputAsWell = false, $options = [])
    {
        $this->setOptions($options);

        if(file_exists($this->options['location'] . $this->options['filename']) && !$this->options['overwrite'])
            throw new \Exception('File exists already');

        foreach($this->mutations as $m)
        {
            $this->parser->addMutation($m->getDate(), $m->amount, $m->balance, $m->getDescription(), $m);
        }

        $mt940 = $this->parser->generate();
        \file_put_contents($this->options['location'] . $this->options['filename'], $mt940);

        if($outputAsWell)
            return $mt940;
        else
        	return true;
        
    }
}
